/*
 * cesard.c : serveur de cryptographie faible par la mthode de Csar :
 * faire tourner les lettres dans l'alphabet.r
 *
 * Ce programme parfaitement inutile est juste un exemple
 * d'implmentation de serveur TCP. On peut facilement s'en inspirer
 * pour implmenter d'autres serveur. Pour cel il faut
 *	- changer les #defines ci-dessous ;
 *	- recrire la fonction answer().
 *
 * Auteur : Edgar Bonet Orozco
 * Inspir du travail de :
 *      Juan Arturo Nolazco Flores
 *      Pavel Moreno Orozco
 *      Brenda Sanchez
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <netdb.h>
#include <signal.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>

/*
 * Pour dfinir le service "cesar", il faut ajouter la ligne
 *  cesar 3591/tcp  dans /etc/services. On peut lui donner un numro
 * de port diffrent de 3591 mais si le numro de port est infrieur 
 * 1024, le dmon ne pourra fonctionner que sous l'UID root ( moins
 * d'tre lanc par inetd). Si le service n'est pas ainsi dfini, le
 * port numro DEF_PORT est pris par dfaut.
 *
 * On peut lancer le serveur  chaque connexion via inetd. Pour cela il
 * faut rajouter dans /etc/inetd.conf la ligne  cesar stream tcp nowait
 * nobody  /usr/local/bin/cesard cesard  puis demander  inetd de
 * relire son fichier de configuration en tapant  killall -HUP inetd .
 * Ceci suppose que le service cesar a t dfini dans /etc/services,
 * que l'utilisateur nobody est dfini dans /etc/passwd et que
 * l'excutable se trouve dans /usr/local/bin/cesard.
 */

#define SERV_NAME "cesar"	/* service tcp fourni */
#define DEF_PORT 3591		/* port par dfaut */
#define LINE_SZ 256		/* taille du tampon de lecture */

/* Intervalle des caractres ASCII imprimables (entre SP et ~) */
#define ST_ASCII ' '			/* dbut de l'intervalle */
#define LEN_ASCII ('~' - ' ' + 1)	/* longueur */

/* Aide pour l'utilisateur */
#define usage() fprintf(stderr, msg_usage, argv[0], SERV_NAME, DEF_PORT)
char msg_usage[] =
"syntaxe : %s [-h] [-a adresse] [-p port] [-s]
  -h : affiche cette aide ;
  -a adresse : prcise l'adresse IP sur laquelle il faut couter ; ce
        peut tre une adresse numrique ou un nom de domaine ; par
	dfaut le serveur coute toutes les adresse possibles ;
  -p port : prcise le port sur lequel il faut se connecter ; le port
        par dfaut est celui attribu au service %s o %d si ce
	service n'est pas dfini ;
  -s : traite une seule connexion sur stdin/stdout ; ce comportement est
	adopt par dfaut si stdin est un socket, l'option -s est donc
	inutile si le serveur est lanc par inetd.\n";

/*
 * Dcale de offset les caractres imprimables
 * de la chane s sans toucher aux autres
 */
void cesar(int offset, char *s)
{
	offset %= LEN_ASCII;
	if (offset < 0) offset += LEN_ASCII;
	for ( ; *s != '\0' ; s++) {
		if (*s < ST_ASCII || *s >= ST_ASCII + LEN_ASCII)
			continue;
		*s = ((*s - ST_ASCII) + offset)
			% LEN_ASCII + ST_ASCII;
	}
}

/*
 * Rpond  une connexion ouverte sur stdin et stdout et ferme la connexion
 */
void answer(void)
{
	char line[LINE_SZ];
	char *pos;
	int offset;
	int offset_ok = 0;

	/* Le tampon stdout doit tre vid pour chaque ligne crite */
	setlinebuf(stdout);

	/* Il faut saluer le client */
	printf("Serveur cesard prt\r\n");

	/* Le client doit demander un dcalage */
	do {
		if (fgets(line, LINE_SZ, stdin) == NULL) {
			fprintf(stderr, "fils : erreur fgets");
			return;
		}
		if (sscanf(line, "offset %d", &offset) == 1)
			offset_ok = 1;
		else
			printf("Prcisez d'abord un dcalage avec "
				"la commande \'offset\'\r\n");
	} while (!offset_ok);

	/* On est prt  traiter le texte */
	printf("entrez votre texte, terminez par \'.\'\r\n");
	while (1) {
		if (fgets(line, LINE_SZ, stdin) == NULL) {
			fprintf(stderr, "fils : erreur fgets");
			break;
		}

		/* Un point en dbut de ligne doit tre supprim... */
		pos = line;
		if (*pos == '.') {
			pos++;

			/* ...mais un point seul sur une ligne marque la
			 * fin de celle-ci */
			if (*pos == '\r' || *pos == '\n') {
				printf("Au revoir\r\n");
				break;
			}
		}

		/* On  crypte  le texte et on le renvoie */
		cesar(offset, pos);
		printf("%s", pos);
	}

	/* Y'a plus qu' fermer la connexion */
	if(fclose(stdout) == EOF)
		perror("fils : fclose");
}

/*
 * La fonction main n'a rien de spcifique  ce serveur. Elle peut tre
 * utilise pour implmenter d'autres serveurs TCP.
 */
int main(int argc, char **argv)
{
	int opt;
	char my_name[LINE_SZ] = "*";
	int my_port = -1;
	struct servent *service;
	struct hostent *resolv;
	struct in_addr my_in_addr;
	struct sockaddr_in my_addr;
	struct sockaddr_in his_addr;
	int addr_size;
	int s, t;
	int child;

	/* Lecture de la ligne de commande */
	while ((opt=getopt(argc, argv, "ha:p:s")) != EOF)
		switch (opt) {
			case 'a' :
				strncpy(my_name, optarg, LINE_SZ-1);
				break;
			case 'p' :
				my_port = atoi(optarg);
				break;
			case 'h' :
				usage();
				return 0;
			case 's' :
				answer();
				return 0;
			default :
				usage();
				return -1;
		}

	/* Si stdin est un socket, il n'y a plus qu' rpondre */
	addr_size = sizeof(struct sockaddr);
	if (getpeername(fileno(stdin), (struct sockaddr *) &his_addr,
		&addr_size) == 0) {
		answer();
		return 0;
	} else if (errno != ENOTSOCK) {
		perror("getpeername");
		return -1;
	}

	/* Recherche de l'adresse IP */
	if (strcmp(my_name, "*") == 0)

		/* Toutes les adresse disponibles */
		my_in_addr.s_addr = htonl(INADDR_ANY);
	else if (strspn(my_name, "0123456789.") == strlen(my_name)) {

		/* Adresse numrique */
		if (inet_aton(my_name, &my_in_addr) == 0) {
			fprintf(stderr, "erreur inet_aton\n");
			return -1;
		}
	} else {

		/* Nom de domaine */
		resolv = gethostbyname(my_name);
		if (resolv == NULL) {
			perror("gethostbyname");
			return -1;
		}
		my_in_addr.s_addr =
			*((unsigned long *) resolv->h_addr_list[0]);
	}

	/* Recherche du numro de port */
	if (my_port == -1) {
		service = getservbyname(SERV_NAME, "tcp");
		if (service != NULL) my_port = service->s_port;
		else my_port = DEF_PORT;
	}

	/* Crer le socket */
	s = socket(AF_INET, SOCK_STREAM, 0);
	if (s == -1) {
		perror("socket");
		return -1;
	}

	/* L'attacher  mon adresse */
	my_addr.sin_family = AF_INET;
	my_addr.sin_port = htons(my_port);
	my_addr.sin_addr = my_in_addr;
	if (bind(s, (struct sockaddr *) &my_addr,
		sizeof(struct sockaddr)) == -1) {
		perror("bind");
		return -1;
	}

	if (listen(s, 5) == -1) {
		perror("listen");
		return -1;
	}

	addr_size = sizeof(struct sockaddr);

	/* viter les zombies : cf. man sigaction */
	signal(SIGCHLD, SIG_IGN);

	/* Boucle d'attente des connexions */
	while (1) {

		/* On prend une connexion */
		if ((t = accept(s, (struct sockaddr *) &his_addr,
			&addr_size)) == -1) {

			/* Interruption reue, on reprend */
			if (errno == EINTR) continue;
			perror("accept");
			return -1;
		}

		/* On cre un fils pour s'en occuper */
		if ((child = fork()) == -1) {
			perror("fork");
			return -1;
		}

		/* Code du fils */
		if (child == 0) {

			/* Fermer le socket qui reste en coute */
			if (close(s) == -1) {
				perror("fils : close socket");
				return -1;
			}

			/* Rediriger les entres/sorties */
			if (close(fileno(stdin)) == -1) {
				perror("fils : close stdin");
				return -1;
			}
			if (dup(t) == -1) {
				perror("fils : dup stdin");
				return -1;
			}
			if (close(fileno(stdout)) == -1) {
				perror("fils : close stout");
				return -1;
			}
			if (dup(t) == -1) {
				perror("fils : dup stout");
				return -1;
			}

			/* Rpondre  la requte */
			answer();
			return 0;
		}

		/* Code du pre */
		if(close(t) == -1) {
			perror("pre : close");
			return -1;
		}
	}
}
